import re
from dataclasses import dataclass


@dataclass
class TAI:
    """International Atomic Time."""

    pass


@dataclass
class UTC:
    """Coordinated Universal Time."""

    pass


@dataclass
class UTCk:
    """Coordinated Universal Time laboratory realization."""

    lab: str


@dataclass
class Custom:
    """Custom reference time system."""

    s: str


ReferenceTimeUnion = TAI | UTC | UTCk | Custom


class ReferenceTime:
    """Reference time system."""

    _system: ReferenceTimeUnion

    def __init__(self, system: ReferenceTimeUnion):
        self._system = system

    def __str__(self):
        match self._system:
            case TAI():
                return "TAI"
            case UTC():
                return "UTC"
            case UTCk(lab):
                return f"UTC({lab})"
            case Custom(s):
                return s

    def __eq__(self, other):
        print(other.__class__)
        print(self.__class__)
        if not isinstance(other, ReferenceTime):
            return False

        print(self)
        if self._system.__class__ == other._system.__class__:
            print(self._system)
            match self._system:
                case UTCk(lab):
                    return lab == other._system.lab
                case Custom(s):
                    return s == other._system.s
                case _:
                    return True

        return False

    @classmethod
    def from_str(cls, s: str):
        if s == "tai":
            return TAI()
        elif s == "utc":
            return UTC()
        elif (lab := re.search(r"UTC\((\s*\w+\s*)\)", s)) is not None:
            return UTCk(lab.group(1).strip())
        else:
            return Custom(s)



from .cggtts import Coordinates, File, load, loads
from .code import Code
from .delay import Delay, SystemDelay
from .reftime import TAI, UTC, Custom, ReferenceTime, UTCk
from .track import CommonViewClass, Track
from .version import Version

__all__ = [
    "load",
    "loads",
    "Code",
    "Coordinates",
    "File",
    "SystemDelay",
    "Delay",
    "ReferenceTime",
    "UTC",
    "UTCk",
    "Custom",
    "TAI",
    "CommonViewClass",
    "Track",
    "Version",
]



from dataclasses import dataclass, field

import numpy as np

from .code import Code


@dataclass
class Delay:
    """GNSS receiver delay.

    Delays are always specified in nanoseconds.
    """

    value: np.float64

    def __format__(self, format_spec):
        return f"{self.delay}"

    @property
    def value_seconds(self):
        return self.value * 1e-9

    def add_value(self, rhs):
        self.value += rhs


@dataclass
class SystemDelay:
    """Describes the total measurement system's delay."""

    cab_delay: np.float64 = 0.0
    ref_delay: np.float64 = 0.0
    delays: dict[Code, Delay] = field(default_factory=dict)
    cal_id: str | None = None

    def total_delay(self, code: Code) -> np.float64 | None:
        """Return the total delay for a given code."""

        if code not in self.delays:
            return None

        return self.cab_delay + self.ref_delay + self.delays[code].value

    def total_delays(self) -> list[tuple[Code, np.float64]]:
        """Group the total delays for each code."""

        return [(code, self.total_delay(code)) for code in self.delays.keys()]



from enum import Enum, auto


class NonSupportedVersionError(Exception):
    """Raised when the version is not supported."""

    pass


class Version(Enum):
    """Supported versions of the CGGTTS format.

    Currently, only version 2E (latest) is supported. We also support reading 30s raw
    data generated by programs like R2CGGTTS.
    """

    VERSION_2E = auto()
    RAW = auto()

    @classmethod
    def from_str(cls, version: str):
        if version == "2E":
            return cls.VERSION_2E
        else:
            raise NonSupportedVersionError(f"Version {version} is not supported.")



import types
from dataclasses import dataclass
from enum import StrEnum
from typing import Iterator

import numpy as np
from hifitime import Duration, Epoch, Unit

TrackType = types.SimpleNamespace()
TrackType.WITH_IONO = 24
TrackType.WITHOUT_IONO = 21
TrackType.RAW = 12


class CommonViewClass(StrEnum):
    """Kind of Common View system."""

    SINGLE = "99"
    MULTI = "FF"


@dataclass
class IonosphericData:
    """Ionospheric data attached to a CGGTTS track.

    Attributes:
        msio: Measured ionospheric delay in nanoseconds.
        smsi: Slope of the measured ionospheric delay in picoseconds per second.
        isg: Root-mean-square of the residuals in nanoseconds.
    """

    msio: np.float64 | None
    smsi: np.float64 | None
    isg: np.float64 | None


@dataclass
class TrackData:
    """Track data.

    All values expressed in seconds.

    Attributes:
        refsv: Difference to the reference SV.
        srsv: Slope of the reference SV.
        refsys: Difference to the reference system.
        srsys: Slope of the reference system.
        dsg: Root-mean-square of the residuals to linear fit.
        ioe: Three-digit decimal code indicating the ephemeris used for the computation.
        mdtr: Modeled tropospheric delay.
        smdt: Slope of the modeled tropospheric delay.
        mdio: Modeled ionospheric delay.
        smdi: Slope of the modeled ionospheric delay.
    """

    refsv: np.float64
    srsv: np.float64 | None
    refsys: np.float64
    srsys: np.float64 | None
    dsg: np.float64 | None
    ioe: np.uint16
    mdtr: np.float64
    smdt: np.float64 | None
    mdio: np.float64
    smdi: np.float64 | None

    @classmethod
    def _parse_data(cls, items: Iterator[str]):
        refsv = np.float64(next(items)) * 0.1e-9
        srsv = np.float64(next(items)) * 0.1e-12
        refsys = np.float64(next(items)) * 0.1e-9
        srsys = np.float64(next(items)) * 0.1e-12
        dsg = np.float64(next(items)) * 0.1e-9
        ioe = np.uint16(next(items))
        mdtr = np.float64(next(items)) * 0.1e-9
        smdt = np.float64(next(items)) * 0.1e-12
        mdio = np.float64(next(items)) * 0.1e-9
        try:
            smdi = np.float64(next(items)) * 0.1e-12
        except ValueError:
            smdi = None

        return cls(refsv, srsv, refsys, srsys, dsg, ioe, mdtr, smdt, mdio, smdi)

    @classmethod
    def parse_without_iono(cls, items: Iterator[str]):
        return cls._parse_data(items), None

    @classmethod
    def parse_with_iono(cls, items: Iterator[str]):
        data = cls._parse_data(items)

        msio = np.float64(next(items)) * 0.1e-9
        try:
            smsi = np.float64(next(items)) * 0.1e-12
        except ValueError:
            smsi = None
        isg = np.float64(next(items)) * 0.1e-9

        iono = IonosphericData(msio, smsi, isg)
        return data, iono

    @classmethod
    def parse_raw(cls, items: Iterator[str]):
        refsv = np.float64(next(items)) * 0.1e-9
        refsys = np.float64(next(items)) * 0.1e-9
        ioe = np.uint16(next(items))
        mdtr = np.float64(next(items)) * 0.1e-9
        mdio = np.float64(next(items)) * 0.1e-9

        try:
            msio = np.float64(next(items)) * 0.1e-9
        except ValueError:
            msio = None

        iono = IonosphericData(msio=msio, smsi=None, isg=None)
        return cls(refsv, None, refsys, None, None, ioe, mdtr, None, mdio, None), iono


@dataclass
class Track:
    """A CGGTTS measurement.

    Attributes:
        cv_class: Common View class (Single/Multi channel).
        epoch: Epoch of this track.
        duration: Tracking duration.
        sv: Satellite vehicle tracked.
        elevation: Elevation at track midpoint, expressed in degrees.
        azimuth: Azimuth at track midpoint, expressed in degrees.
        data: Track data.
        iono: Optional ionospheric compensation terms.
        fr: GLONASS channel frequency [1:24], 0 for other GNSS.
        hc: Hardware/receiver channel [0:99], 0 if unknown.
        frc:  Carrier frequency standard 3 letter code.
    """

    cv_class: CommonViewClass | None
    epoch: Epoch
    duration: Duration | None
    sv: str
    elevation: np.float64
    azimuth: np.float64
    data: TrackData
    iono: IonosphericData | None
    fr: np.uint8 | None
    hc: np.uint8 | None
    frc: str

    @classmethod
    def from_str(cls, s: str):
        items = s.strip().split()

        num_items = len(items)
        it_items = iter(items)

        sv = next(it_items)
        cv_class = (
            CommonViewClass(next(it_items)) if num_items != TrackType.RAW else None
        )
        mjd = np.int32(next(it_items))

        sttime = next(it_items)
        if len(sttime) < 6:
            raise ValueError("Invalid start time format")

        hh = np.uint8(sttime[0:2])
        mm = np.uint8(sttime[2:4])
        ss = np.uint8(sttime[4:6])

        epoch = Epoch.init_from_mjd_utc(np.float64(mjd))
        epoch = epoch + Unit.Hour * np.float64(hh)
        epoch = epoch + Unit.Minute * np.float64(mm)
        epoch = epoch + Unit.Second * np.float64(ss)

        if num_items != TrackType.RAW:
            duration = Duration.from_total_nanoseconds(
                (np.float64(next(it_items)) * 1e9).astype(np.int64)
            )
        else:
            duration = None

        elevation = np.float64(next(it_items)) * 0.1
        azimuth = np.float64(next(it_items)) * 0.1

        data: TrackData
        iono: IonosphericData
        match num_items:
            case TrackType.WITH_IONO:
                data, iono = TrackData.parse_with_iono(it_items)
            case TrackType.WITHOUT_IONO:
                data, iono = TrackData.parse_without_iono(it_items)
            case TrackType.RAW:
                data, iono = TrackData.parse_raw(it_items)
            case _:
                raise ValueError("Invalid number of items")

        if num_items != TrackType.RAW:
            fr = np.uint8(next(it_items))
            hc = np.uint8(next(it_items))
        else:
            fr = None
            hc = None

        frc = next(it_items)

        # TODO: Process checksum
        if num_items != TrackType.RAW:
            _ = next(it_items)

        return cls(
            cv_class, epoch, duration, sv, elevation, azimuth, data, iono, fr, hc, frc
        )



import re
from dataclasses import dataclass, field
from typing import BinaryIO

import numpy as np
from hifitime import Epoch, TimeScale

from .code import Code
from .delay import SystemDelay
from .reftime import UTC, ReferenceTime
from .track import Track
from .version import Version

CURRENT_VERSION: str = "2E"


@dataclass
class Coordinates:
    """Coordinates"""

    x: np.float64 = 0.0
    y: np.float64 = 0.0
    z: np.float64 = 0.0

    def __str__(self):
        return f"{self.x} {self.y} {self.z}"


@dataclass
class File:
    """A CGGTTS data structure.

    Contains a list of comparisons between a local clock and a reference time system.

    Attributes:
        version: CGGTTS version used in this file. We only support version 2E (latest)
        release_date: The release date of the CGGTTS file.
        station: The station name.
        receiver: GNSS receiver info.
        num_channels: The number of GNSS receiver channels.
        ims: Ionospheric Measurement System used, if any.
        reference_time: The reference time system.
        reference_frame: The reference frame, coordinates system and conversions.
        apc_coordinates: Antenna phase center coordinates in meters.
        comments: Comments, if any.
        delay: System delay.
        tracks: List of succesive measurements.
    """

    version: Version = Version.VERSION_2E
    release_date: Epoch | None = field(
        default_factory=Epoch.init_from_gregorian_at_midnight(
            2014, 2, 20, TimeScale.UTC
        )
    )
    station: str = "LAB"
    receiver: str | None = None
    num_channels: np.uint16 = 0
    ims: str | None = None
    reference_time: ReferenceTime = field(default_factory=ReferenceTime(UTC()))
    reference_frame: str | None = None
    apc_coordinates: Coordinates = field(default_factory=Coordinates())
    comments: str | None = None
    delay: SystemDelay = field(default_factory=SystemDelay())
    tracks: list[Track] = field(default_factory=list)


def load(fp: BinaryIO) -> File:
    """Load a CGGTTS file from a binary stream.

    Args:
        BinaryIO: A binary stream.

    Returns:
        A CGGTTS file.
    """
    b = fp.read()

    try:
        s = b.decode()
    except AttributeError:
        raise TypeError(
            "File must be opened in binary mode, e.g. use `open('file', 'r')`"
        ) from None

    return loads(s)


def loads(s: str) -> File:
    lines = iter(s.splitlines())

    # Initialize variables
    release_date = None
    num_channels = 0
    receiver = None
    ims = None
    station = "LAB"
    comments = None
    system_delay = SystemDelay()
    reference_time = ReferenceTime(UTC())
    reference_frame = None
    apc_coordinates = Coordinates()

    # Version always comes first
    line = next(lines)
    if line.strip() == "RAW CLOCK RESULTS":
        version = Version.RAW
    else:
        m = re.match(r"CGGTTS\s+GENERIC DATA FORMAT VERSION = (\w+)", line)
        match m:
            case None:
                raise ValueError("Version format error.")
            case _:
                version = Version.from_str(m.group(1))

    # Parse the header
    while True:
        line = next(lines)

        if line.startswith("REV DATE = "):
            m = re.match(r"REV DATE = (\d{4})-(\d{2})-(\d{2})", line)
            match m:
                case None:
                    raise ValueError("Release date format error.")
                case _:
                    release_date = Epoch.init_from_gregorian_at_midnight(
                        int(m.group(1)),
                        int(m.group(2)),
                        int(m.group(3)),
                        TimeScale.UTC,
                    )
        elif line.startswith("RCVR = "):
            # TODO: Correctly parse receiver info
            m = re.match(r"RCVR = (\w+)", line)
            match m:
                case None:
                    continue
                case _:
                    receiver = m.group(1)
        elif line.startswith("CH = "):
            m = re.match(r"CH = (\d+)", line)
            match m:
                case None:
                    continue
                case _:
                    num_channels = np.uint16(m.group(1))
        elif line.startswith("IMS = "):
            m = re.match(r"IMS = (\w+) (\w+) (\w+) (\d+) (\w+)", line)
            match m:
                case None:
                    continue
                case _:
                    ims = m.group(1)
        elif line.startswith("LAB = "):
            match line.strip("LAB = "):
                case "":
                    continue
                case s:
                    station = s.strip()
        elif line.startswith("X = "):
            m = re.match(r"X =\s+([+-]?\d+\.\d+)", line)
            match m:
                case None:
                    continue
                case _:
                    apc_coordinates.x = np.float64(m.group(1))
        elif line.startswith("Y = "):
            m = re.match(r"Y =\s+([+-]?\d+\.\d+)", line)
            match m:
                case None:
                    continue
                case _:
                    apc_coordinates.y = np.float64(m.group(1))
        elif line.startswith("Z = "):
            m = re.match(r"Z =\s+([+-]?\d+\.\d+)", line)
            match m:
                case None:
                    continue
                case _:
                    apc_coordinates.z = np.float64(m.group(1))
        elif line.startswith("FRAME = "):
            f = line[7:].strip()
            if f != "?":
                reference_frame = f
        elif line.startswith("COMMENTS = "):
            c = line.strip("COMMENTS = ").strip()
            if c and c != "NO COMMENTS":
                comments = c
        elif line.startswith("REF = "):
            m = re.match(r"REF = (.+)", line)
            match m:
                case None:
                    continue
                case _:
                    reference_time = ReferenceTime.from_str(m.group(1))
        elif "DLY = " in line:
            items = line.split()
            dual_carrier = "," in line

            if len(items) < 4:  #  Format mismatch
                continue

            match items[0]:
                case "CAB":
                    system_delay.cab_delay = np.float64(items[3])
                case "REF":
                    system_delay.ref_delay = np.float64(items[3])
                case "SYS":
                    if "CAL_ID" in line:
                        offset = line.rfind("=")
                        cal_id = line[offset + 1 :].strip()
                        if cal_id != "NA":
                            system_delay.cal_id = cal_id

                    if dual_carrier:
                        value = np.float64(items[3])
                        code = items[6].replace("),", "")

                        # TODO: Iterate over all codes
                    elif value := np.float64(items[3]):
                        code = items[6].replace("),", "")
                        if code := Code[code]:
                            system_delay.delays[code] = value
                case "INT":
                    if "CAL_ID" in line:
                        offset = line.rfind("=")
                        cal_id = line[offset + 1 :].strip()
                        if cal_id and cal_id != "NA":
                            system_delay.cal_id = cal_id

                    if dual_carrier:
                        value = np.float64(items[3])
                        code = items[6].replace("),", "")

                        # TODO: Iterate over all codes
                    elif value := np.float64(items[3]):
                        code = items[6].replace("),", "")
                        if code := Code[code]:
                            system_delay.delays[code] = value
                case _:
                    continue
        elif line.startswith("CKSUM = "):
            # TODO: Process checksum
            break

    # Skip lines (blank, labels, units)
    next(lines)
    next(lines)
    next(lines)

    # Parse tracks
    tracks = []
    while True:
        try:
            line = next(lines)
            if line == "":  # We are done parsing
                break

            tracks.append(Track.from_str(line))
        except StopIteration:  # EOF
            break

    return File(
        version=version,
        release_date=release_date,
        station=station,
        receiver=receiver,
        num_channels=num_channels,
        ims=ims,
        reference_time=reference_time,
        reference_frame=reference_frame,
        apc_coordinates=apc_coordinates,
        comments=comments,
        delay=system_delay,
        tracks=tracks,
    )



from enum import StrEnum, auto


class Code(StrEnum):
    """GNSS code types."""

    C1 = auto()
    C2 = auto()
    P1 = auto()
    P2 = auto()
    E1 = auto()
    E5 = auto()
    B1 = auto()
    B2 = auto()
